from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any, Generic, ParamSpec, TypeVar, overload

from loguru import logger
from pydantic import BaseModel, ConfigDict

from .resource_pool import ResourcePool

InputT = ParamSpec("InputT")
RetT = TypeVar("RetT")


class TaskProcessError(Exception):
    """Exception raised when a task process fails."""


class Task(BaseModel, Generic[InputT, RetT]):
    """
    Task to be executed by the workflow.

    Example:
    >>> my_task = Task(name="my_task", fn=lambda: print("Hello, world!"))
    >>> my_task()
    Hello, world!
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str
    fn: Callable[InputT, RetT]
    retry_interval: list[int] = [10, 30, 60]
    resource_pool: ResourcePool[Any] | None = None
    resource_modifier: Callable[[Callable[InputT, RetT], Any], Callable[InputT, RetT]] | None = None

    def __call__(self, *args: InputT.args, **kwargs: InputT.kwargs) -> RetT:
        if self.resource_pool is not None or self.resource_modifier is not None:
            logger.warning(
                f"Task `{self.name}` has a resource pool or resource modifier, which is not supported in sync mode"
            )
        return self.fn(*args, **kwargs)

    def submit(self, *args: InputT.args, **kwargs: InputT.kwargs) -> asyncio.Task[RetT]:
        retry_interval = self.retry_interval[:]

        async def wrapper_fn() -> RetT:
            try:
                if self.resource_pool is not None:
                    resource = await self.resource_pool.acquire()
                    logger.info(f"Acquired resource by task [blue]{self.name}[/blue]: {resource}")
                    if self.resource_modifier is not None:
                        fn = self.resource_modifier(self.fn, resource)
                    else:
                        fn = self.fn
                    try:
                        return await asyncio.to_thread(fn, *args, **kwargs)
                    finally:
                        self.resource_pool.release(resource)
                        logger.info(f"Released resource: {resource}")
                else:
                    return await asyncio.to_thread(self.fn, *args, **kwargs)
            except TaskProcessError as e:
                logger.error(f"Failed to execute task: {e}")
                if retry_interval:
                    retry = retry_interval.pop(0)
                    logger.info(f"Retry task `{self.name}` after {retry} seconds")
                    await asyncio.sleep(retry)
                    return await wrapper_fn()
                raise e

        return asyncio.create_task(wrapper_fn())


@overload
def task(fn: Callable[InputT, RetT]) -> Task[InputT, RetT]: ...


@overload
def task(
    *,
    name: str | None = None,
    resource_pool: ResourcePool[Any] | None = ...,
    resource_modifier: Callable[[Callable[InputT, RetT], Any], Callable[InputT, RetT]] | None = None,
) -> Callable[[Callable[InputT, RetT]], Task[InputT, RetT]]: ...


def task(
    fn: Callable[InputT, RetT] | None = None,
    *,
    name: str | None = None,
    resource_pool: ResourcePool[Any] | None = None,
    resource_modifier: Callable[[Callable[InputT, RetT], Any], Callable[InputT, RetT]] | None = None,
) -> Callable[[Callable[InputT, RetT]], Task[InputT, RetT]] | Task[InputT, RetT]:
    """Decorator to create a task.

    Example:
    >>> @task
    >>> def my_task(a: int, b: int) -> int:
    >>>     return a + b
    >>> my_task.name
    'my_task'
    >>> @task(name="custom_name")
    >>> def my_task(a: int, b: int) -> int:
    >>>     return a + b
    >>> my_task.name
    'custom_name'
    """

    def decorator(fn: Callable[InputT, RetT]) -> Task[InputT, RetT]:
        return Task(
            name=name or getattr(fn, "__name__", "unnamed"),
            fn=fn,
            resource_pool=resource_pool,
            resource_modifier=resource_modifier,
        )

    if fn is None:
        return decorator
    return decorator(fn)



from __future__ import annotations

import pytest


@pytest.mark.benchmark
@pytest.mark.parametrize(
    "config_path",
    ["./examples/simple.toml", "./examples/matrix.toml"],
)
def test_cli(config_path: str):
    from typer.testing import CliRunner

    from nanoflow.__main__ import app

    runner = CliRunner()
    result = runner.invoke(app, [config_path])
    assert result.exit_code == 0, f"Exit code was {result.exit_code}, expected 0. Error: {result.exc_info}"


def test_cli_error():
    from typer.testing import CliRunner

    from nanoflow.__main__ import app

    runner = CliRunner()
    result = runner.invoke(app, ["./examples/error.toml"])
    assert result.exit_code != 0


@pytest.mark.asyncio
async def test_tui():
    import toml

    from nanoflow.config import WorkflowConfig
    from nanoflow.tui import Nanoflow

    workflow_config = WorkflowConfig.model_validate(toml.load("./examples/simple.toml"))
    app = Nanoflow(workflow_config)
    async with app.run_test() as pilot:
        await pilot.press("d")
        await pilot.press("f1")
        assert app.is_running is True
        await pilot.press("q")
        assert app.is_running is False



from __future__ import annotations

from collections.abc import Generator
from itertools import product
from typing import Any, Literal

from pydantic import BaseModel


class DefaultDict(dict):
    def __missing__(self, key: str):
        return f"{{{key}}}"


def flatten_matrix(matrix: dict[str, list[str]]) -> Generator[dict[str, str], Any, None]:
    matrix_keys, matrix_values = zip(*matrix.items(), strict=True)
    product_matrix = product(*matrix_values)
    for values in product_matrix:
        yield DefaultDict(zip(matrix_keys, values, strict=True))


class TaskConfig(BaseModel):
    """Task config.

    Example:
    >>> config = TaskConfig(command="echo", args=["task1"])
    >>> config.get_command()
    'echo task1'
    >>> config = TaskConfig(command="echo", args=["{task}"])
    >>> config.format({"task": "task1"}).get_command()
    'echo task1'
    >>> config.get_command()
    'echo {task}'
    >>> config.format({"task": "task1"}, inplace=True)
    >>> config.get_command()
    'echo task1'
    >>> config = TaskConfig(command="echo", args=["{task}"], matrix={"task": ["task1", "task2"]})
    >>> for name, task in config.wrap_matrix("task:{task}").items():
    ...     print(name, task.get_command())
    task:task1 echo task1
    task:task2 echo task2
    """

    command: str
    matrix: dict[str, list[str]] | None = None
    args: list[str] = []
    deps: list[str] = []

    def get_command(self) -> str:
        assert self.matrix is None, "Matrix is not None, you must run wrap_matrix first"
        return f"{self.command} {' '.join(self.args)}"

    def format(
        self, template_values: dict[str, str], *, format_deps: bool = False, inplace: bool = False
    ) -> TaskConfig:
        if inplace:
            task = self
        else:
            task = self.model_copy(deep=True)

        task.command = task.command.format_map(template_values)
        task.args = [arg.format_map(template_values) for arg in task.args]
        if format_deps:
            task.deps = [dep.format_map(template_values) for dep in task.deps]
        return task

    def wrap_matrix(self, name: str) -> dict[str, TaskConfig]:
        assert self.matrix is not None, "You cannot run wrap_matrix without matrix"
        tasks: dict[str, TaskConfig] = {}
        for i, template_values in enumerate(flatten_matrix(self.matrix)):
            task = self.format(template_values, inplace=False)
            task.matrix = None
            task_name = name.format(**template_values)
            if task_name in tasks:
                task_name = f"{task_name}_{i}"
            tasks[task_name] = task
        return tasks


class WorkflowConfig(BaseModel):
    """
    Workflow config.

    Example:
    >>> config = WorkflowConfig(
    >>>     name="test",
    >>>     resources="gpus",
    >>>     tasks={
    >>>         "task1": TaskConfig(command="echo 'task1'"),
    >>>         "task2": TaskConfig(command="echo 'task2'", deps=["task1"]),
    >>>         "task3": TaskConfig(command="echo 'task3'", deps=["task2"]),
    >>>     }
    >>> )
    >>> config.to_nodes()
    {'task1': [], 'task2': ['task1'], 'task3': ['task2']}
    >>> config = WorkflowConfig(
    >>>     name="test",
    >>>     matrix={"a": ["1", "2"]},
    >>>     tasks={
    >>>         "task1": TaskConfig(command="echo 'task{a}'"),
    >>>         "task2": TaskConfig(command="echo 'task{a}'", deps=["task1"]),
    >>>     }
    >>> )
    >>> config.to_nodes()
    {'0_task1': [], '0_task2': ['0_task1'], '1_task1': [], '1_task2': ['1_task1']}
    """

    name: str
    tasks: dict[str, TaskConfig]
    matrix: dict[str, list[str]] | None = None
    resources: Literal["gpus"] | list[str] | None = None

    def model_post_init(self, __context: Any) -> None:
        if self.matrix is None:
            return

        tasks: dict[str, TaskConfig] = {}
        for i, template_values in enumerate(flatten_matrix(self.matrix)):
            for task_name, task_config in self.tasks.items():
                if task_config.matrix is not None:
                    wrapped_tasks = task_config.wrap_matrix(task_name)
                    for wrapped_name, wrapped_task in wrapped_tasks.items():
                        wrapped_task_name = f"{i}_{wrapped_name}"
                        task = wrapped_task.format(template_values, inplace=False)
                        task.deps = [f"{i}_{dep}" for dep in task.deps]
                        tasks[wrapped_task_name] = task
                else:
                    task_name = f"{i}_{task_name}"
                    task = task_config.format(template_values, inplace=False)
                    task.deps = [f"{i}_{dep}" for dep in task.deps]
                    tasks[task_name] = task

        self.tasks = tasks

    def to_nodes(self) -> dict[str, list[str]]:
        nodes: dict[str, list[str]] = {}
        for task_name, task_config in self.tasks.items():
            nodes[task_name] = task_config.deps

        return nodes



from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine
from typing import Any, Generic, ParamSpec, overload

from pydantic import BaseModel

P = ParamSpec("P")


class Workflow(BaseModel, Generic[P]):
    name: str
    fn: Callable[P, Coroutine[Any, Any, None]]

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        await self.fn(*args, **kwargs)

    def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
        asyncio.run(self.fn(*args, **kwargs))


@overload
def workflow(fn: Callable[P, Coroutine[Any, Any, None]]) -> Workflow[P]: ...


@overload
def workflow(*, name: str | None = ...) -> Callable[[Callable[P, Coroutine[Any, Any, None]]], Workflow[P]]: ...


def workflow(
    fn: Callable[P, Coroutine[Any, Any, None]] | None = None, *, name: str | None = None
) -> Callable[[Callable[P, Coroutine[Any, Any, None]]], Workflow[P]] | Workflow[P]:
    def decorator(fn: Callable[P, Coroutine[Any, Any, None]]) -> Workflow[P]:
        return Workflow(name=name or fn.__name__, fn=fn)

    if fn is None:
        return decorator
    return decorator(fn)



from __future__ import annotations

import os
import subprocess
from collections.abc import Callable

import networkx as nx

from .resource_pool import ResourcePool, UnlimitedPool
from .task import Task, TaskProcessError, task


@task
def layer_nodes(node_dependencies: dict[str, list[str]]) -> list[list[str]]:
    graph = nx.DiGraph()

    for node, dependencies in node_dependencies.items():
        graph.add_node(node)
        for dependency in dependencies:
            graph.add_edge(dependency, node)

    level = {}
    for node in nx.topological_sort(graph):
        level[node] = max((level[pred] + 1 for pred in graph.predecessors(node)), default=0)

    max_level = max(level.values(), default=-1)
    parallel_nodes = [[] for _ in range(max_level + 1)]
    for node, lvl in level.items():
        parallel_nodes[lvl].append(node)

    return parallel_nodes


@task
def get_available_gpus(threshold: float = 0.05) -> list[str]:
    gpu_info: str = subprocess.check_output(
        ["nvidia-smi", "--query-gpu=memory.used,memory.total", "--format=csv,nounits,noheader"]
    ).decode("utf-8")
    gpus = [tuple(map(int, line.split(","))) for line in gpu_info.strip().split("\n")]

    free_gpus = []
    for i, (used, total) in enumerate(gpus):
        usage_ratio = used / total
        if usage_ratio <= threshold:
            free_gpus.append(str(i))

    return free_gpus


def create_command(
    name: str,
    command: str,
    *,
    update_hook: Callable[[str, bytes], None] | None = None,
    environ: dict[str, str] | None = None,
) -> Callable[[], None]:
    def inner_fn() -> None:
        if update_hook is not None:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=environ)
            assert process.stdout is not None
            for line in process.stdout:
                update_hook(name, line)
        else:
            process = subprocess.Popen(command, shell=True, env=environ)
        returncode = process.wait()
        if returncode != 0:
            raise TaskProcessError(f"Task `{name}` failed with return code {returncode}")

    return inner_fn


def create_gpu_task(
    name: str,
    command: str,
    *,
    pool: ResourcePool,
    update_hook: Callable[[str, bytes], None] | None = None,
) -> Task[[], None]:
    def set_visible_gpu(fn: Callable[[], None], resource: int) -> Callable[[], None]:
        # TODO: To support custom resources, we need to set the resource in the environ
        environ = os.environ.copy()
        environ["CUDA_VISIBLE_DEVICES"] = str(resource)
        environ["FORCE_COLOR"] = "1"
        return create_command(name, command, update_hook=update_hook, environ=environ)

    return task(name=name, resource_pool=pool, resource_modifier=set_visible_gpu)(lambda: None)


def create_task(
    name: str,
    command: str,
    *,
    pool: ResourcePool | None = None,
    update_hook: Callable[[str, bytes], None] | None = None,
) -> Task[[], None]:
    def set_base_environ(fn: Callable[[], None], resource: int) -> Callable[[], None]:
        # TODO: To support custom resources, we need to set the resource in the environ
        environ = os.environ.copy()
        environ["FORCE_COLOR"] = "1"
        return create_command(name, command, update_hook=update_hook, environ=environ)

    if pool is None:
        pool = UnlimitedPool(None)
    return Task(
        name=name,
        fn=lambda: None,
        resource_pool=pool,
        resource_modifier=set_base_environ,
    )



from __future__ import annotations

import asyncio
from collections.abc import Callable

from loguru import logger

from .config import WorkflowConfig
from .resource_pool import ResourcePool
from .utils import create_gpu_task, create_task, get_available_gpus, layer_nodes
from .workflow import Workflow


class Executor:
    def __init__(self, workflow: Workflow):
        # TODO: support multiple workflows
        self.workflow = workflow

    @classmethod
    def from_configs(
        cls,
        config: WorkflowConfig,
        update_hook: Callable[[str, bytes], None] | None = None,
    ) -> Executor:
        logger.info("Creating GPU resource pool and parallel tasks")
        layered_nodes = layer_nodes(config.to_nodes())
        resources = config.resources
        if resources == "gpus":
            resource_pool = ResourcePool(get_available_gpus())
            layered_tasks = [
                [
                    create_gpu_task(node, config.tasks[node].get_command(), pool=resource_pool, update_hook=update_hook)
                    for node in nodes
                ]
                for nodes in layered_nodes
            ]
        else:
            if resources is not None:
                logger.warning("Use of custom resources is experimental and may not work as expected")
                resource_pool = ResourcePool(resources)
            else:
                resource_pool = None
            layered_tasks = [
                [
                    create_task(node, config.tasks[node].get_command(), pool=resource_pool, update_hook=update_hook)
                    for node in nodes
                ]
                for nodes in layered_nodes
            ]

        async def workflow_fn():
            for tasks in layered_tasks:
                start_time = asyncio.get_event_loop().time()
                logger.info(f"Starting execution of {len(tasks)} tasks")
                await asyncio.gather(*[task.submit() for task in tasks])
                end_time = asyncio.get_event_loop().time()
                logger.info(f"Execution completed, actual time taken: {end_time - start_time:.2f} seconds")

        workflow = Workflow(name=config.name, fn=workflow_fn)
        return cls(workflow)

    def run(self):
        self.workflow.run()



from __future__ import annotations

import asyncio
from logging import Handler
from pathlib import Path
from typing import Literal

import toml
import typer
from loguru import logger
from rich.highlighter import NullHighlighter
from rich.logging import RichHandler

from nanoflow import WorkflowConfig
from nanoflow.executor import Executor

app = typer.Typer()


def init_logger(log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], handler: Handler):
    logger.remove()
    logger.add(handler, format="{message}", level=log_level)


@app.command()
def run(config_path: Path, use_tui: bool = False):
    handler = RichHandler(highlighter=NullHighlighter(), markup=True)
    init_logger("DEBUG", handler)
    workflow_config = WorkflowConfig.model_validate(toml.load(config_path))
    if use_tui:  # pragma: no cover
        from nanoflow.tui import Nanoflow

        app = Nanoflow(workflow_config)
        executor = Executor.from_configs(workflow_config, update_hook=app.update_log)

        async def start():
            await asyncio.gather(executor.workflow(), app.run_async())

        asyncio.run(start())
    else:
        executor = Executor.from_configs(workflow_config)
        executor.run()



from __future__ import annotations

from typing import ClassVar

from textual.app import App, Binding, BindingType, ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Footer, Label, Markdown, RichLog, TabbedContent, Tabs

from .config import WorkflowConfig


class HelpScreen(ModalScreen[None]):
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("q", "app.quit", "Quit", show=False),
        Binding("escape,f1,?", "app.pop_screen()", "Close help", key_display="esc"),
    ]

    HELP_MARKDOWN = """\
### How do I quit the app?

Press `Ctrl+C` on your keyboard.
`q` also works if an input isn't currently focused.
"""

    def compose(self) -> ComposeResult:
        with Vertical(id="help-container") as vertical:
            vertical.border_title = "Help"
            with VerticalScroll():
                yield Markdown(self.HELP_MARKDOWN, id="help-markdown")
            yield Markdown(
                "Use `pageup`, `pagedown`, `up`, and `down` to scroll.",
                id="help-scroll-keys-info",
            )
        yield Footer()


class Nanoflow(App[None]):
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("d", "toggle_dark", "Toggle dark mode"),
        Binding("q", "app.quit", "Quit", show=False),
        Binding("f1,?", "help", "Help"),
    ]

    def __init__(self, workflow_config: WorkflowConfig):
        super().__init__()
        self.workflow_config = workflow_config

    def update_log(self, task_name: str, line: bytes):
        if self.is_running:
            self.query_one(f"#log-{task_name}", RichLog).write(line.decode())

    def compose(self) -> ComposeResult:
        with TabbedContent(*list(self.workflow_config.tasks.keys())):
            for task_name in self.workflow_config.tasks:
                yield RichLog(id=f"log-{task_name}")
        yield Footer()

    def on_mount(self) -> None:
        """Focus the tabs when the app starts."""
        self.query_one(Tabs).focus()

    def on_tabs_tab_activated(self, event: Tabs.TabActivated) -> None:
        """Handle TabActivated message sent by Tabs."""
        label = self.query_one(Label)
        label.visible = True
        label.update(event.tab.label)

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark

    async def action_help(self) -> None:
        if isinstance(self.screen, HelpScreen):
            self.pop_screen()
        else:
            await self.push_screen(HelpScreen())



from __future__ import annotations

import asyncio
from typing import Generic, TypeVar

T = TypeVar("T")


class ResourcePool(Generic[T]):
    def __init__(self, resources: list[T]):
        self.resources = {res: asyncio.Semaphore(1) for res in resources}

    async def acquire(self) -> T:
        while True:
            for res, sem in self.resources.items():
                if sem.locked():
                    continue
                if await sem.acquire():
                    return res
            await asyncio.sleep(0.1)

    def release(self, res: T):
        self.resources[res].release()


class UnlimitedPool(ResourcePool[T]):
    def __init__(self, resource: T):
        self.resource = resource

    async def acquire(self) -> T:
        return self.resource

    def release(self, res: T):
        pass
